import React, { useEffect, useRef, useState } from 'react';

type Props = {
  children: React.ReactNode;
  className?: string;
  style?: React.CSSProperties;
  horizontal: boolean;
  divisorSize: number;
  defaultProportion?: number;
  classNameLeft?: string;
  classNameRight?: string;
};

function convertRemToPixels(rem: number) {
  return rem * parseFloat(getComputedStyle(document.documentElement).fontSize);
}

export const Resizable = ({
  children,
  className,
  style,
  horizontal,
  divisorSize,
  defaultProportion,
  classNameLeft,
  classNameRight,
  ...props
}: Props) => {
  const ref = useRef<HTMLDivElement>(null);
  const children2 = children as React.ReactElement[];
  const [firstSize, setFirstSize] = React.useState<number>(0);
  const [secondSize, setSecondSize] = React.useState<number>(0);
  const [dragging, setDragging] = React.useState<boolean>(false);
  const [windowSize, setWindowSize] = useState({ width: window.innerWidth, height: window.innerHeight });

  useEffect(() => {
    const handleResize = () => {
      setWindowSize({ width: window.innerWidth, height: window.innerHeight });
    };

    window.addEventListener('resize', handleResize);

    return () => {
      window.removeEventListener('resize', handleResize);
    };
  }, []);

  useEffect(() => {
    if (ref.current) {
      const rect = ref.current.getBoundingClientRect();
      const _defaultProportion = defaultProportion || 0.5;
      if (horizontal) {
        setFirstSize(rect.width * _defaultProportion - divisorSize);
        setSecondSize(rect.width * (1 / _defaultProportion) - divisorSize);
      } else {
        setFirstSize(rect.height * _defaultProportion - divisorSize);
        setSecondSize(rect.height * (1 / _defaultProportion) - divisorSize);
      }
    }
  }, [ref.current, windowSize]);

  function onMouseDown(e: React.MouseEvent<HTMLDivElement, MouseEvent>) {
    setDragging(true);
    window.addEventListener('mouseup', onMouseUp);
  }
  const onMouseUp = () => {
    setDragging(false);
    window.removeEventListener('mouseup', onMouseUp);
  };

  function onMouseMove(e: React.MouseEvent<HTMLDivElement, MouseEvent>) {
    if (dragging) {
      if (ref.current) {
        const rect = ref.current.getBoundingClientRect();
        const relativeX = e.clientX - rect.left;
        const relativeY = e.clientY - rect.top;
        const minSizeX = 16;
        const minSizeY = 28;

        if (horizontal) {
          setFirstSize(Math.max(minSizeX, relativeX));
          setSecondSize(Math.max(minSizeX, rect.width - relativeX));
        } else {
          setFirstSize(Math.max(minSizeY, relativeY));
          setSecondSize(Math.max(minSizeY, rect.height - relativeY));
        }
      }
    }
  }
  function onTouchStart(e: React.TouchEvent<HTMLDivElement>) {
    setDragging(true);
  }
  function onTouchMove(e: React.TouchEvent<HTMLDivElement>) {}
  function onTouchEnd(e: React.TouchEvent<HTMLDivElement>) {
    setDragging(false);
  }
  function onTouchCancel(e: React.TouchEvent<HTMLDivElement>) {
    setDragging(false);
  }

  if (!children2[0]) return children2[1];
  if (!children2[1]) return children2[0];

  return (
    <>
      {dragging && <div className="absolute top-0 left-0 w-screen h-screen z-10 cursor-grabbing" onMouseMove={onMouseMove}></div>}
      <div
        className={`w-full h-full flex flex-grow ${horizontal ? 'flex-row' : 'flex-col'} ${className !== undefined ? className : ''} `}
        style={style}
        {...props}
        ref={ref}
      >
        {firstSize > 0 && (
          <>
            <div
              className={'h-full w-full' + (classNameLeft !== undefined ? classNameLeft : '')}
              style={horizontal ? { maxWidth: firstSize } : { maxHeight: firstSize }}
            >
              {children2[0]}
            </div>
            <div
              className={' ' + (horizontal ? 'cursor-col-resize' : 'cursor-row-resize')}
              style={horizontal ? { minWidth: divisorSize } : { minHeight: divisorSize }}
              onMouseDown={onMouseDown}
              onTouchStart={onTouchStart}
              onTouchMove={onTouchMove}
              onTouchEnd={onTouchEnd}
              onTouchCancel={onTouchCancel}
            ></div>
            <div
              className={'h-full w-full ' + (classNameRight !== undefined ? classNameRight : '')}
              style={horizontal ? { maxWidth: secondSize } : { maxHeight: secondSize }}
            >
              {children2[1]}
            </div>
          </>
        )}
      </div>
    </>
  );
};
