import { createSlice, PayloadAction } from '@reduxjs/toolkit';
import type { RootState } from './store';
import { getDimension, extractStatistics, GraphMetaData } from '../statistics';

export interface GraphQueryResultFromBackendPayload {
  queryID: string;
  result: {
    payload: GraphQueryResultFromBackend;
    type: string;
  };
}

export type NodeAttributes = { [key: string]: unknown };

export interface GraphQueryResultFromBackend {
  nodes: {
    id: string;
    label?: string;
    attributes: NodeAttributes;
  }[];

  edges: {
    id: string;
    label?: string;
    attributes: NodeAttributes;
    from: string;
    to: string;
  }[];

  // TODO: Also include type in node and edge
  // TODO: The backend should send all the different entitytypes and relationtypes in the result
}

export interface Node {
  id: string;
  label: string;
  attributes: NodeAttributes;
  mldata?: any; // FIXME
  /* type: string[]; */
}
export interface Edge {
  attributes: NodeAttributes;
  from: string;
  to: string;
  id: string;
  label: string;
  /* type: string; */
}

// Define a type for the slice state
export interface GraphQueryResult {
  metaData: GraphMetaData;
  nodes: Node[];
  edges: Edge[];
  queryingBackend: boolean;
}

// Define the initial state using that type
export const initialState: GraphQueryResult = {
  metaData: { nodes: { labels: [], types: {} }, edges: { labels: [], types: {} } },
  nodes: [],
  edges: [],
  queryingBackend: false,
};

export const graphQueryResultSlice = createSlice({
  name: 'graphQueryResult',
  initialState,
  reducers: {
    assignNewGraphQueryResult: (state, action: PayloadAction<GraphQueryResultFromBackendPayload>) => {
      const payload = action.payload.result.payload;

      // Only keep one node and one edge per id. This is also done in the backend, but we do it here as well to be sure.
      const nodeIDs = new Set(payload.nodes.map((node) => node.id));
      const edgeIDs = new Set(payload.edges.map((edge) => edge.id));
      let nodes = [...nodeIDs].map((nodeID) => payload.nodes.find((node) => node.id === nodeID) as unknown as Node);
      let edges = [...edgeIDs].map((edgeID) => payload.edges.find((edge) => edge.id === edgeID) as unknown as Edge);

      const metaData: GraphMetaData = {
        nodes: { labels: [], count: nodes.length, types: {} },
        edges: { labels: [], count: edges.length, types: {} },
      };

      nodes = nodes.map((node) => {
        let _node = { ...node };
        // TODO!: only works for neo4j
        let nodeType: string = node.id.split('/')[0];
        let innerLabels = node?.attributes?.labels as string[];

        if (innerLabels?.length > 0) nodeType = innerLabels[0] as string;
        if (!metaData.nodes.labels.includes(nodeType)) metaData.nodes.labels?.push(nodeType);
        if (!metaData.nodes.types[nodeType]) metaData.nodes.types[nodeType] = { count: 0, attributes: {} };
        metaData.nodes.types[nodeType].count++;

        Object.entries(_node.attributes).forEach(([attributeId, attributeValue]) => {
          if (!metaData.nodes.types[nodeType].attributes[attributeId]) {
            const dimension = getDimension(attributeValue);
            metaData.nodes.types[nodeType].attributes[attributeId] = { dimension, values: [], statistics: {} };
          }
          metaData.nodes.types[nodeType].attributes[attributeId].values?.push(attributeValue);
        });

        _node.label = nodeType;
        return _node;
      });

      edges = edges.map((edge) => {
        let _edge = { ...edge };
        // TODO!: only works for neo4j
        let edgeType: string = edge.id.split('/')[0];

        if (!_edge.id.includes('/')) edgeType = edge.attributes.Type as string;
        if (!metaData.edges.labels?.includes(edgeType)) metaData.edges.labels?.push(edgeType);
        if (!metaData.edges.types[edgeType]) metaData.edges.types[edgeType] = { count: 0, attributes: {} };
        metaData.edges.types[edgeType].count++;

        Object.entries(_edge.attributes).forEach(([attributeId, attributeValue]) => {
          if (!metaData.edges.types[edgeType].attributes[attributeId]) {
            const dimension = getDimension(attributeValue);
            metaData.edges.types[edgeType].attributes[attributeId] = { dimension, values: [], statistics: {} };
          }
          metaData.edges.types[edgeType].attributes[attributeId].values?.push(attributeValue);
        });

        _edge.label = edgeType;
        return _edge;
      });

      // Assign new state
      state.metaData = extractStatistics(metaData);
      state.nodes = nodes;
      state.edges = edges;
      state.queryingBackend = false;
    },
    resetGraphQueryResults: (state) => {
      // Assign new state
      state.metaData = { nodes: { labels: [], types: {} }, edges: { labels: [], types: {} } };
      state.nodes = [];
      state.edges = [];
      state.queryingBackend = false;
    },
    queryingBackend: (state) => {
      state.queryingBackend = true;
    },
  },
});

export const { assignNewGraphQueryResult, resetGraphQueryResults, queryingBackend } = graphQueryResultSlice.actions;

// Other code such as selectors can use the imported `RootState` type
export const selectGraphQueryResult = (state: RootState) => state.graphQueryResult;
export const selectGraphQueryResultNodes = (state: RootState) => state.graphQueryResult.nodes;
export const selectGraphQueryResultLinks = (state: RootState) => state.graphQueryResult.edges;
export const selectGraphQueryResultMetaData = (state: RootState) => state.graphQueryResult.metaData;

export default graphQueryResultSlice.reducer;
