/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

/* istanbul ignore file */
/* The comment above was added so the code coverage wouldn't count this file towards code coverage.
 * We do not test components/renderfunctions/styling files.
 * See testing plan for more details.*/
import React, { useState } from 'react';
import { Node, Handle, Position, NodeProps } from 'reactflow';
import styles from './entity.module.scss';
import { SchemaReactflowNodeWithFunctions } from '../../../model/reactflow';
import { QueryElementTypes } from '@graphpolaris/shared/lib/querybuilder';
import { SchemaEntityPopup } from './SchemaEntityPopup';
import { Popup } from '@graphpolaris/shared/lib/components/Popup';
import { SchemaNode } from '../../../model';

/**
 * EntityNode is the node that represents the database entities.
 * It has four handles on the top, bottom, right and left to connect edges.
 * It also has a text to display the entity name and show the amount of attributes it contains.
 * @param {NodeProps} param0 The data of an entity flow element.
 */
export const EntityNode = React.memo(({ id, selected, data }: NodeProps<SchemaReactflowNodeWithFunctions>) => {
  const [openPopup, setOpenPopup] = useState(false);

  /**
   * adds drag functionality in order to be able to drag the entityNode to the schema
   * @param event React Mouse drag event
   */
  const onDragStart = (event: React.DragEvent<HTMLDivElement>) => {
    const eventData: SchemaNode = {
      type: QueryElementTypes.Entity,
      name: id,
      attributes: data.attributes.map((attribute) => attribute.attributes).flat(), // TODO: this seems wrong
    };
    event.dataTransfer.setData('application/reactflow', JSON.stringify(eventData));
    event.dataTransfer.effectAllowed = 'move';
  };

  /**
   * displays the NodeQualityPopup when clicked, and hides them when clicked again
   */
  const onClickToggleNodeQualityPopup = (): void => {
    data.toggleNodeQualityPopup(id);
  };

  /**
   * displays the attribute analytics popup menu when clicked, and hides them when clicked again
   */
  const onClickToggleAttributeAnalyticsPopupMenu = (): void => {
    data.toggleAttributeAnalyticsPopupMenu(id);
  };

  return (
    <>
      {openPopup && (
        <Popup open={openPopup} hAnchor="left" className="-top-8" offset="-20rem">
          <SchemaEntityPopup data={data} onClose={() => setOpenPopup(false)} />
        </Popup>
      )}
      <div
        className={`rounded-sm transition-all duration-150 shadow shadow-dark/10 hover:shadow-md hover:shadow-dark/20 min-w-[4rem] max-w-[8rem] text-[0.8rem] bg-gradient-to-r pt-1 from-[#FFA952] to-[#D66700]`}
        onDragStart={(event) => onDragStart(event)}
        onDragStartCapture={(event) => onDragStart(event)}
        onMouseDownCapture={(event) => {
          if (!event.shiftKey) event.stopPropagation();
        }}
        onClickCapture={(event) => {
          setOpenPopup(!openPopup);
        }}
        draggable
      >
        <div className={`py-1 ${selected ? 'bg-secondary-300' : 'bg-secondary-100'}`}>
          <Handle
            style={{ pointerEvents: 'none' }}
            id="entityTargetLeft"
            position={Position.Left}
            className={styles.handleTriangleLeft}
            type="target"
          ></Handle>
          <Handle
            style={{ pointerEvents: 'none' }}
            id="entitySourceRight"
            position={Position.Right}
            className={styles.handleTriangleRight}
            type="source"
          ></Handle>
          <div style={{ pointerEvents: 'none' }} className="p-2 py-1 text-center truncate">
            <span className="">{id}</span>
          </div>
        </div>
      </div>
    </>
  );
});

EntityNode.displayName = 'EntityNode';
